"""Constant definitions.

This file is loaded during the program initialisation before the main module is
imported. Consequently, this file must not import any other modules, or those
modules will be initialised before the main module, which means that the DEBUG
value may not have been set correctly.

This module is intended to be imported with 'from ... import *' semantics but
it does not provide an __all__ specification.

"""

#: Enable debug features. Should never be changed manually but is set to True
#: automatically when running `test_game.py`.
DEBUG = False
DEBUG_POSLOG = False

#: Version string. This may be represented somewhere in-game. It is also read
#: by `setup.py` as part of its versioning features.
VERSION = u"1.0"

#: The directory (relative to the top level) wherein all the resources for the
#: game are stored, probably subdivided into types of resources. See `data.py`.
DATA_DIR = "data"

#: The name of the game used in locating the saved settings directory. Its best
#: not to have any spaces in this name.
CONFIG_NAME = "Loopback"

#: The caption that appears at the top of the window. Obviously this is only
#: visible in windowed mode.
CAPTION = u"Loopback"

#: The "top-level" tick rate; the maximum number of times per second that the
#: controller will call its tick method.
TICK_RATE = 60.0

#: The "top-level" update rate; the maximum number of times per second that the
#: controller will call its update method.
UPDATE_RATE = 60.0

NUMBER_OF_TIMES = 9

FONT_NAME = "Silkscreen"

MENU_FONT_OPTS = dict(
    font_name = FONT_NAME,
    font_size = 0.04,
)

FIRING_RATE = 4
EXTRA_PROJECTILE_RADIUS = 4
BASE_SPEED = 1.0
SLOWDOWN = 1.0

CORE_HP = 80
ATTACK_ZONE_SIZE = 20

WAVE_TICKS = 45.0 * TICK_RATE

MUSICS = [
    "CheeZeeLab.mp3",
]
OTHER_MUSIC = "Sunshine.mp3",

FADENESS = 0.8
FADERATE = 0.1
ENDING_TICKS = 45


INTRO_WAVE_1_TICKS = 2.5 * TICK_RATE
INTRO_WAVE_2_TICKS_A = 0.5 * TICK_RATE
INTRO_WAVE_2_TICKS_B = 2.0 * TICK_RATE
INTRO_ENDING_TICKS = 4.0 * TICK_RATE


LASER_SEGMENT_LENGTH = 10

ALERT_TIME = 40

LEVEL_BRIGHTNESS = 127
