from __future__ import division

from pyglet import graphics
from pyglet.gl import *
from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED
from pyglet.window import key
from pyglet import text

import random # for lightning
import math

import interface
import mode
from vector import v
import menumode

import level
import sound
import data
import texts

from entity import Enemy, PlayerShip, Projectile, Fragment
from gamedata import GameData

import config
from constants import *



class OrderTextureGroup(graphics.TextureGroup):
    def __init__(self, order, texture, parent=None):
        super(OrderTextureGroup, self).__init__(texture, parent)
        self.order = order
    def __cmp__(self, other):
        if isinstance(other, OrderTextureGroup):
            return cmp(self.order, other.order)
        return -1
    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
            self.order == other.order and
            self.texture.target == other.texture.target and
            self.texture.id == other.texture.id and
            self.parent == other.parent)

class World(interface.panel.Panel):

    world_height = interface.attribute.Attribute(480.0)

    def draw_content(self):
        cw, ch = self.content_size
        scale = ch / self.world_height
        glPushMatrix()
        #glTranslatef(cw / 2.0, ch / 2.0, 0.0)
        glScalef(scale, scale, 1.0)
        self.draw_world()
        glPopMatrix()

    def draw_world(self):
        pass

class GameMode(mode.Mode):
    name = "game"

    def __init__(self, gamedata=None):
        super(GameMode, self).__init__()

        ## Interface
        ############

        self.root = interface.content.Root(
                aspect=(4, 3),
                scissor = True,
                background_color = (0.1, 0.1, 0.1, 1.0),
        )

        self.world = self.root.add(World)
        self.world.draw_world = self.draw_world

        self.score_display = self.root.add(interface.content.Text,
                text = u"000000",
                font_name = FONT_NAME,
                font_size = 0.06,
                valign = "top",
        )

        self.wave_fade = self.root.add(interface.content.Color,
                color = (0.0, 0.0, 0.0, 0.0),
                width = 1.0,
                height = 0.0,
                valign = "bottom",
                halign = "left",
        )

        self.wave_display = self.root.add(interface.content.Text,
                text = u"Wave 1/9",
                font_name = FONT_NAME,
                font_size = 0.06,
                anchor_x = "right",
                halign = "left",
                valign = 0.85,
                width = 1.0,
                color = (1.0, 1.0, 1.0, 1.0),
        )

        self.timer = self.root.add(interface.content.Text,
                text = u"00:00",
                font_name = FONT_NAME,
                font_size = 0.06,
                valign = "bottom",
        )

        self.fade = self.root.add(interface.content.Color,
                color = (0.0, 0.0, 0.0, 0.0),
                visible = False,
        )

        self.handler = interface.event.Handler(self.root)

        ## Interface: Pause Menu
        ########################

        self.pause = self.root.add(interface.content.Color,
                color = (0.0, 0.0, 0.0, 0.0),
                visible = False,
        )

        self.pause_menu = self.pause.add(menumode.TextMenu,
                item_options = MENU_FONT_OPTS,
        )

        self.pause_menu.add_item("Resume", self.do_resume)
        self.pause_menu.add_item("Restart Wave", self.do_retry)
        self.pause_menu.add_item("Quit", self.do_quit)
        self.pause_menu.change_selection(0)

        ## Interface: Replay Menu
        #########################

        self.replay_pause = self.root.add(interface.content.Color,
                color = (0.0, 0.0, 0.0, 0.0),
                visible = False,
        )

        self.replay_menu = self.replay_pause.add(menumode.TextMenu,
                item_options = MENU_FONT_OPTS,
        )

        self.replay_menu.add_item("Resume", self.do_resume)
        self.replay_menu.add_item("Restart Replay", self.do_replay)
        self.replay_menu.add_item("Retry Wave", self.do_retry)
        self.replay_menu.add_item("Quit", self.do_quit)
        self.replay_menu.change_selection(0)

        ## Interface: Pass Menu
        #######################

        self.pass_ = self.root.add(interface.content.Color,
                color = (0.0, 0.0, 0.0, 0.0),
                width = 1.1,
                height = 1.1,
                visible = False,
        )

        self.pass_title = self.pass_.add(interface.content.Text,
                font_name = FONT_NAME,
                font_size = 0.08,
                text = u"Wave clear!",
                valign = "top",
        )

        self.pass_menu = self.pass_.add(menumode.TextMenu,
                item_options = MENU_FONT_OPTS,
        )

        self.pass_menu.add_item("Continue", self.do_continue)
        self.pass_menu.add_item("Show Replay", self.do_replay)
        self.pass_menu.change_selection(0)

        self.pass_text = self.pass_.add(interface.content.Text,
                font_name = FONT_NAME,
                font_size = 0.02,
                text = u"",
                valign = "bottom",
                margin = 0.02,
        )

        ## Interface: Fail Menu
        #######################

        self.fail = self.root.add(interface.content.Color,
                color = (0.0, 0.0, 0.0, 0.0),
                width = 1.1,
                height = 1.1,
                visible = False,
        )

        self.fail_title = self.fail.add(interface.content.Text,
                font_name = FONT_NAME,
                font_size = 0.08,
                text = u"Core destroyed!",
                valign = "top",
        )

        self.fail_menu = self.fail.add(menumode.TextMenu,
                item_options = MENU_FONT_OPTS,
        )

        self.fail_menu.add_item("Retry Wave", self.do_retry)
        self.fail_menu.add_item("Show Replay", self.do_replay)
        self.fail_menu.add_item("Quit", self.do_quit)
        self.fail_menu.change_selection(0)

        self.fail_text = self.fail.add(interface.content.Text,
                font_name = FONT_NAME,
                font_size = 0.02,
                text = u"",
                valign = "bottom",
                margin = 0.02,
        )

        ## Interface: Win Menu
        ######################

        self.win = self.root.add(interface.content.Color,
                color = (0.0, 0.0, 0.0, 0.0),
                width = 1.1,
                height = 1.1,
                visible = False,
        )

        self.win_title = self.win.add(interface.content.Text,
                font_name = FONT_NAME,
                font_size = 0.08,
                text = u"You Win!",
                valign = "top",
        )

        self.win_menu = self.win.add(menumode.TextMenu,
                item_options = MENU_FONT_OPTS,
        )

        self.win_menu.add_item("Show Replay", self.do_replay)
        self.win_menu.add_item("Quit", self.do_quit)
        self.win_menu.change_selection(0)

        self.win_text = self.win.add(interface.content.Text,
                font_name = FONT_NAME,
                font_size = 0.04,
                text = u"Difficulty: %s\n Score: %s",
                valign = "bottom",
        )

        ## Graphics
        ###########

        self.batch = graphics.Batch()
        self.real_batch = graphics.Batch()
        self.level_vlist = None
        self.core_vlist = None
        self.wave_fade_vlist = None

        ## Other
        ########

        self.is_paused = False
        self.load_level(gamedata)
        self.state = None
        self.fade_call = None
        self.fade_target = 0.0
        self.hists = {}

    def __del__(self):
        if self.level_vlist:
            self.level_vlist.delete()
            self.core_vlist.delete()

    ## Level Stuff
    ##############

    def cleanup_level(self):
        for e in self.enemies:
            e.destroy()
        for s in self.ships:
            s.destroy()

    def load_level(self, gamedata, replay=False):
        if hasattr(self, "level"):
            self.cleanup_level()
        self.level = level.Level()
        if gamedata is None:
            self.gamedata = GameData()
        else:
            self.gamedata = gamedata
        self.ticks = 0
        self.ships = map(lambda p: PlayerShip(self, p), gamedata.previous)
        if not replay:
            s1 = PlayerShip(self)
            s1.k = self.keys # do this better
            self.ships.append(s1)
        
        for i, s in enumerate(self.ships):
            pos, heading = self.gamedata.starts[i]
            s.teleport(pos)
            s.heading = heading
            s.echo = not s.recording
            s.start()
        
        self.enemies = []
        self.projectiles = []
        self.spawners = self.gamedata.get_spawners(self, replay)
        mul = BASE_SPEED
        for i in xrange(len(self.ships) - 1, -1, -1):
            self.ships[i].rate = mul
            self.spawners[i].rate = mul
            mul *= SLOWDOWN
        self.level.core.hp = CORE_HP
        self.lasers = []
        self.alerts = []
        self.is_ending = False
        self.ending_ticks = None
        self.wave_display.text = "Wave %d/%d\nDefend The Core!" % (len(self.ships), NUMBER_OF_TIMES)
        
        self.score = self.gamedata.fixedscore

    ## Mode Methods
    ###############

    def connect(self, control):
        super(GameMode, self).connect(control)
        self.window.push_handlers(self.handler)
        self.window.push_handlers(self.keys)
        self.root.window = self.window

    def disconnect(self):
        self.window.remove_handlers(self.keys)
        self.window.pop_handlers()
        super(GameMode, self).disconnect()


    ## State Changes
    ################

    def set_state(self, state):
        self.pause.visible = False
        self.replay_pause.visible = False
        self.pass_.visible = False
        self.fail.visible = False
        self.win.visible = False
        self.is_paused = True
        if state == "game" or state == "replay":
            self.is_paused = False
        elif state == "pause":
            self.pause.visible = True
            self.pause_menu.change_selection(0)
        elif state == "replaypause":
            self.replay_pause.visible = True
            self.replay_menu.change_selection(0)
        elif state == "pass":
            self.pass_text.text = texts.pass_texts[len(self.ships)]
            self.pass_.visible = True
            self.pass_menu.change_selection(0)
        elif state == "fail":
            self.fail_text.text = texts.fail_texts[len(self.ships)]
            self.fail.visible = True
            self.fail_menu.change_selection(0)
        elif state == "win":
            self.win.visible = True
            self.win_menu.change_selection(0)
        self.state = state

    def do_pause(self):
        self.set_state("pause")

    def do_replaypause(self):
        self.set_state("replaypause")

    def do_pass(self):
        if self.gamedata.won:
            if self.score > config.highscore:
                config.save_option("highscore", self.score)
                self.win_title.text = "New High Score!"
            self.win_text.text = u"Difficulty: %s\nScore: %06d" % (self.gamedata.diff, self.score)
            self.set_state("win")
        else:
            self.set_state("pass")

    def do_fail(self):
        if self.gamedata.wave == 1:
            if self.fail_menu.children[1].text == "Show Replay":
                self.fail_menu.remove(self.fail_menu.children[1])
        elif len(self.fail_menu.children) == 2:
            self.fail_menu.add_item("Show Replay", self.do_replay)
            self.fail_menu._children.insert(1,self.fail_menu._children.pop())
        self.set_state("fail")

    def do_resume(self):
        self.set_state("game")

    def do_continue(self):
        if self.gamedata.won:
            self.set_state("win")
        else:
            self.load_level(self.gamedata)
            self.set_state("game")

    def do_retry(self):
        self.load_level(self.gamedata)
        self.set_state("game")

    def do_replay(self):
        if self.state == "pass":
            self.replay_menu.children[2].text = "Next Wave"
        if self.state == "fail":
            self.replay_menu.children[2].text = "Retry Wave"
        if self.state == "win":
            self.replay_menu.remove(self.replay_menu.children[2])
        self.load_level(self.gamedata, True)
        self.set_state("replay")

    def do_quit(self):
        self.control.switch_handler("menu")

    def start_fade(self, func):
        self.is_paused = True
        self.fade_call = func
        self.fade_target = FADENESS
        self.fade.visible = True
        self.fade.color = (0.0, 0.0, 0.0, 0.0)

    ## Event Handlers
    #################

    def on_key_press(self, symbol, modifiers):
        if symbol == key.ESCAPE:
            if self.state == "pause" or self.state == "replaypause":
                self.do_resume()
            elif not self.is_ending:
                if self.state == "replay":
                    self.start_fade(self.do_replaypause)
                else:
                    self.start_fade(self.do_pause)
        elif DEBUG and symbol == key.P:
            self.ticks = WAVE_TICKS - 1
        elif DEBUG and symbol == key.F:
            self.level.core.hp = 0
        else:
            return EVENT_UNHANDLED
        return EVENT_HANDLED

    def tick(self):

        # Process fade state.
        if self.fade_call is not None:
            alpha = self.fade.color[3]
            alpha = min(alpha + FADERATE, self.fade_target)
            self.fade.color = (0.0, 0.0, 0.0, alpha)
            if alpha == self.fade_target:
                self.fade_call()
                self.fade_call = None
                self.fade_target = 0.0

        # Stop if paused.
        if self.is_paused:
            return

        # Reverse the fade.
        if self.fade.color[3] > 0.0:
            alpha = self.fade.color[3]
            alpha = max(alpha - FADERATE, 0.0)
            self.fade.color = (0.0, 0.0, 0.0, alpha)
        if self.fade.color[3] > 0.0:
            return

        # Update game.
        self.new_enemies = []

        for s in self.ships:
            s.tick()

        pi = level.ProjectileIndex()
        for p in self.projectiles:
            p.tick()
            if p.damage > 0:
                pi.add(p)

        self.lasers = []
        self.alerts = []
        for e in self.enemies:
            e.tick()
            if (not e.echo) and e.ticks <= ALERT_TIME:
                self.alerts.append(e)
            for p in pi.colliders(e):
                p.dead = True
                e.damage(p.damage, not p.echo, p.heading)
            if not e.dead and e.can_attack() and not self.is_ending:
                sound.underattack()
                self.level.core.hp -= e.dps
                ap = self.level.attack_point(e.pos)
                vec = e.pos - ap
                self.lasers.append((e.pos, ap - vec.safe_scaled_to(10)))
                self.add_fragment(e.rate, ap, vec.angle)

        for s in self.spawners:
            s.tick()
            self.new_enemies.extend(s.new_enemies())

        for cls, p, rs, echo in self.new_enemies:
            e = cls(self, rs)
            e.teleport(p)
            e.rate = s.rate
            e.echo = echo
            self.enemies.append(e)
            e.start()

        self.enemies = [e for e in self.enemies if not e.dead]
        self.projectiles = [p for p in self.projectiles if not p.dead]

        self.ticks += 1

        # Update core color.
        import colorsys
        val = (1 + math.sin(math.pi * self.ticks / 80)) / 2
        pct = max(0, self.level.core.hp / CORE_HP)
        r, g, b = colorsys.hsv_to_rgb(0.33 * pct, 0.8, 0.6+0.3*val)
        if pct == 0:
            color2 = [0, 0, 0, 0]
        else:
            color2 = [int(255 * r), int(255 * g), int(255 * b), 255]
        if self.core_vlist:
            self.core_vlist.colors[:] = 4 * color2

        # Update score.
        self.score_display.text = u"%06d" % self.score

        # Update timer.
        time = max(0, WAVE_TICKS - self.ticks) / float(TICK_RATE)
        seconds = int(time)
        hundredths = int(100 * (time - seconds))
        self.timer.text = u"%02d:%02d" % (seconds, hundredths)

        # Check for FAIL!
        if self.level.core.hp <= 0:
            if not self.is_ending:
                self.core_detonate() # do this *before* setting is_ending
                self.is_ending = True
                self.ending_ticks = ENDING_TICKS
            else:
                self.ending_ticks -= 1
            if self.ending_ticks == 0:
                self.start_fade(self.do_fail)

        # Check for PASS!
        elif self.ticks >= WAVE_TICKS:
            corepos = v((self.level.core.xm, self.level.core.ym))
            for e in self.enemies:
                e.damage(e.hp, False, (e.pos - corepos).angle, silent=True)
            if not self.is_ending:
                self.is_ending = True
                self.ending_ticks = ENDING_TICKS
                sound.win()
            else:
                self.ending_ticks -= 1
            if self.ending_ticks == 0:
                if DEBUG_POSLOG:
                    self.hists[self.gamedata.wave] = h = []
                    for s in self.ships:
                        h.append((s.history, s.rec.history))
                    f = open("lib/log.py", "w")
                    f.write(open("log.py").read())
                    f.write(repr(self.hists))
                self.ships[-1].rec.stop(self.ships[-1].ticks)
                self.gamedata.previous.append(self.ships[-1].rec)
                self.gamedata.starts.append((self.ships[-1].pos, self.ships[-1].heading))
                self.gamedata.fixedscore = self.score
                self.start_fade(self.do_pass)

    def on_draw(self):
        if not self.level_vlist:
            self.make_level_vlist()
        self.window.clear()

        # Update wave display.
        f = lambda t: max(0, min(1, 32*(t-.5)**5+.5))
        self.wave_display.xoffset = f(0.01 * self.ticks) * \
                (self.root.content_size[0] + self.wave_display.content_size[0])
        f = lambda t: max(0, min(1, 1-64*(t-0.5)**6))
        self.wave_fade.height = self.wave_display.content_size[1] / self.root.content_size[1]
        self.wave_fade.yoffset = self.wave_display.content_offset[1]
        self.wave_fade.color = [0, 0, 0, .8*f(0.01 * self.ticks)]

        self.root.draw()

    def make_level_vlist(self):
        tex = data.load_texture("box.png")
        tc = tex.tex_coords
        td = [tc[i] for i in (0, 1, 3, 4, 6, 7, 9, 10)]
        vdata = []
        cdata = []
        tdata = []
        for b in self.level.blocks:
            if b == self.level.core: continue
            o = -0.1
            vdata.extend([b.x1+o, b.y1+o, b.x2-o, b.y1+o, b.x2-o, b.y2-o, b.x1+o, b.y2-o])
            cdata.extend([LEVEL_BRIGHTNESS, LEVEL_BRIGHTNESS, LEVEL_BRIGHTNESS, 255] * 4)
            tdata.extend(td)
        group = OrderTextureGroup(1, tex)
        self.level_vlist = self.batch.add(len(vdata)//2, GL_QUADS,
                group, ('v2f', vdata), ('t2f', tdata), ('c4B', cdata))

        tex = data.load_texture("core.png")
        tc = tex.tex_coords
        td = [tc[i] for i in (0, 1, 3, 4, 6, 7, 9, 10)]
        b = self.level.core
        vdata = [b.x1, b.y1, b.x2, b.y1, b.x2, b.y2, b.x1, b.y2]
        cdata = [255] * 16
        tdata = td
        group = OrderTextureGroup(3, tex)
        self.core_vlist = self.batch.add(len(vdata)//2, GL_QUADS,
                group, ('v2f', vdata), ('t2f', tdata), ('c4B', cdata))

    def draw_world(self):

        self.batch.draw()

        glLineWidth(self.root.content_size[1] / 120)
        glColor3f(1, 1, 0)
        glBegin(GL_LINES)
        for l, ap in self.lasers:
            lvec = l - ap
            segs = 1 + int(lvec.length / LASER_SEGMENT_LENGTH)
            lseg = lvec / segs / 2
            glVertex2f(*ap)
            for ii in xrange(2, segs * 2, 2):
                d = lseg.rotated(random.randrange(360))
                glVertex2f(*(ap + lseg*ii + d))
                glVertex2f(*(ap + lseg*ii + d))
            glVertex2f(*l)
        glEnd()

        glPointSize(self.root.content_size[1] / 180)
        glBegin(GL_POINTS)
        for p in self.projectiles:
            glColor4f(*p.color)
            glVertex2f(*p.pos)
        glEnd()

        glLineWidth(self.root.content_size[1] / 240)
        for e in self.alerts:
            s = 2 + 0.5 * math.sin(e.ticks / 10)
            glColor4f(1, 0, 0, min(35, max(0, 45 - e.ticks)) / 35)
            glPushMatrix()
            glTranslatef(e.pos[0], e.pos[1], 0)
            glScalef(e.radius, e.radius, 1)
            glBegin(GL_LINE_LOOP)
            for idx in xrange(8):
                angle = 2 * math.pi * idx / 8
                glVertex2f(s * math.cos(angle),
                           s * math.sin(angle))
            glEnd()
            glPopMatrix()

    def add_projectile(self, rate, pos, heading, echo, cls=Projectile):
        if not self.is_ending:
            p = cls(self, heading)
            p.rate = rate
            p.teleport(pos)
            p.echo = echo
            self.projectiles.append(p)

    def add_fragment(self, rate, pos, heading):
        self.add_projectile(rate, pos, heading, False, Fragment)

    def core_detonate(self):
        c = self.level.core
        for ii in xrange(500):
            x = random.uniform(c.x1, c.x2)
            y = random.uniform(c.y1, c.y2)
            self.add_fragment(1.0, v((x, y)), random.randrange(360))
        sound.explosion()

    def score_points(self, amount):
        self.score += amount
