import pyglet
import data
import config

class FadingPlayer(pyglet.media.Player):

    def __init__(self):
        super(FadingPlayer, self).__init__()
        self.volume = 0.0
        self.fading = False
        self.fade_time = 0.0
        self.fade_rate = 0.0
        self.fade_grace = 0.0

    def start_fade(self, target, time, grace):
        self.fading = True
        self.fade_time = time
        self.fade_grace = grace
        target = min(1.0, max(0.0, target))
        delta = float(target - self.volume)
        self.fade_rate = delta / time

    def stop_fade(self, dt):
        if self.fade_grace > 0.0:
            self.fade_grace = max(0.0, self.fade_grace - dt)
        if self.fade_grace == 0.0:
            self.fading = False
            self.fade_rate = 0.0

    def update(self, dt):
        if self.fade_time > 0.0:
            self.fade_time = max(0.0, self.fade_time - dt)
            volume = self.volume + self.fade_rate * dt
            self.volume = min(1.0, max(0.0, volume))
        if self.fade_time == 0.0:
            self.stop_fade(dt)

class MusicManager(object):

    def __init__(self):
        self.player = FadingPlayer()
        self.player.eos_action = self.player.EOS_PAUSE
        self.fade_grace = 0.5
        self.orders = []

    def change_song(self, song, fadeout=None, fadein=None):
        if pyglet.media.have_avbin:
            source = data.load_song(song)
            self.orders.append([self.do_fadeout, fadeout])
            self.orders.append([self.do_switch_song, source])
            self.orders.append([self.do_fadein, fadein])
    def random_play(self, music):
        import random
        self.random_songs = list(music)
        random.shuffle(self.random_songs)
        self.orders.append([self.do_random])
    def stop(self):
        self.orders = [[self.do_fadeout, 1.0], [self.do_switch_song, None]]

    def do_random(self):
        if not self.player.source:
            song = self.random_songs.pop(0)
            self.change_song(song, fadein=1.0)
            self.random_songs.append(song)
        self.orders.append([self.do_random])
    def do_fadeout(self, time):
        if time is None:
            self.player.volume = 0.0
        if self.player.volume > 0.0:
            self.player.start_fade(0.0, time, self.fade_grace)
    def do_fadein(self, time):
        if time is None:
            self.player.volume = 1.0
        if self.player.volume < 1.0:
            self.player.start_fade(1.0, time, self.fade_grace)
    def do_switch_song(self, source):
        self.player.next()
        if source is not None:
            if config.sound:
                self.player.queue(source)
                self.player.play()
        else:
            self.player.pause()

    def update(self, dt):
        if not self.player.fading:
            if self.orders:
                order = self.orders.pop(0)
                order[0](*order[1:])
        self.player.update(dt)
