from __future__ import division

import random

import config
import entity

from constants import *

from vector import v

class Spawner(entity.TimeShiftObject):
    def __init__(self, handler, n0, seed):
        self.h = handler
        
        # self.n is the 1-indexed wave number (so code is intuitively readable)
        self.n = (n0 + config.start_wave) % NUMBER_OF_TIMES
        if self.n == 0:
            self.n = NUMBER_OF_TIMES
            
        self.rng = random.Random(seed)
        self.new = []
        self.seed = 0
        
        self.echo = (self.n < handler.gamedata.wave)
        
        self.count = 0
        
    def realtick(self):
        d = self.h.gamedata.difficulty
        p = self.ticks / WAVE_TICKS
    
        if p >= 1:
            return

        if self.n == 1:        

            # the start of waves 1 and 2 is scripted to draw the player to opposite
            # sides of the map, so that he can see his past self kill the past enemies

            # the rest of wave 1 contains reds, then blues, then a mix of both

            if self.ticks == INTRO_WAVE_1_TICKS:
                self.add_group_at_pos(entity.Bumbler, 3, self.h.level.spawns[5])
            elif self.ticks >= INTRO_ENDING_TICKS:
                if (p < 1/3 or p > 2/3) and self.ticks % int(2000 / d) == 0:
                    self.add_group(entity.Bumbler, 3)
                if p > 1/3 and self.ticks % int(1500 / d) == 0:
                    self.add_group(entity.BumblingSeeker, 3)
                
        elif self.n == 2:

            # after the scripted opening, wave 2 consists of alternating groups of blues
            # and reds, with a few single cyans at the end

            if self.ticks == INTRO_WAVE_2_TICKS_A:
                self.add_group_at_pos(entity.Bumbler, 3, self.h.level.spawns[1])
            if self.ticks == INTRO_WAVE_2_TICKS_B:
                self.add_group_at_pos(entity.Bumbler, 3, self.h.level.spawns[2])
            elif self.ticks >= INTRO_ENDING_TICKS:
                x = (self.ticks - 10) % int(4000 / d) 
                if x == 0:
                    self.add_group(entity.BumblingSeeker, 3)
                if x == int(2000 / d):
                    self.add_group(entity.Bumbler, 3)
                    if p > 0.75:
                        self.add_group(entity.Seeker, 1)

        elif self.n == 3:

            # wave 3 consists of groups of 4 greens which enter two at a time from
            # opposite sides of the play area. cyans come in pairs at the end

            if (self.ticks - 10) % int(3600 / d) == 0:
                n = self.rng.randrange(4)
                self.add_group_at_pos(entity.Robot, 4, self.h.level.spawns[n])
                self.add_group_at_pos(entity.Robot, 4, self.h.level.spawns[n + 4])
                if p > 0.75:
                    self.add_group(entity.Seeker, 2)

        elif self.n == 4:

            # wave 4 consists of a random mix of blues, reds and purples, plus a
            # ring of pinks halfway through, and yellows during the second half

            if (self.ticks - 30) % int(2000 / d) == 0:
                cls = self.rng.choice([entity.Bumbler, entity.BumblingSeeker, entity.Whirly])
                self.add_group(cls, 3)
            if self.ticks == int(WAVE_TICKS / 2):
                num = [0] * 8
                for ii in xrange(d):
                    num[self.rng.randrange(8)] += 1
                for jj in xrange(8):    
                    self.add_group_at_pos(entity.Darter, num[jj], self.h.level.spawns[jj])
            if p > 0.5 and self.ticks % int(5000 / d) == 0:
                self.add_group(entity.Breeder, 1)

        elif self.n == 5:

            # wave 5 contains a constant slow influx of yellows in the first half,
            # then groups of five purples in the second

            if p < 0.5 and (self.ticks - 40) % int(3500 / d) == 0:
                self.add_group(entity.Breeder, 1)
            if p > 0.5 and self.ticks % int(2000 / d) == 0:
                self.add_group(entity.Whirly, 5)

        elif self.n == 6:
        
            # wave 6 consists of a steady stream of cyans. simples!

            if (self.ticks - 50) % int(700 / d) == 0:
                self.add_group(entity.Seeker, 1)
        
        elif self.n == 7:

            # wave 7 introduces lilacs, in gradually increasing numbers. blues and
            # purples also come to the party

            if (self.ticks - 60) % int(2000 / d) == 0:
                self.add_group(entity.Sniper, 1 + int(p * 3))
            if self.ticks % int(3000 / d) == 0:
                cls = self.rng.choice([entity.Whirly, entity.BumblingSeeker])
                self.add_group(cls, 2)
        
        elif self.n == 8:

            # wave 8 reprises the fun of previous waves!

            # the yellow assault from the start of wave 5
            if p < 1/3 and (self.ticks - 40) % int(3000 / d) == 0:
                self.add_group(entity.Breeder, 1)

            # additional lilacs from wave 7
            if p >1/3 and p < 2/3 and (self.ticks - 60) % int(2000 / d) == 0:
                self.add_group(entity.Sniper, 1 + int(p * 4))

            # the ring of pinks halfway through wave 4
            if self.ticks == int(WAVE_TICKS / 2):
                num = [0] * 8
                for ii in xrange(d):
                    num[self.rng.randrange(8)] += 1
                for jj in xrange(8):    
                    self.add_group_at_pos(entity.Darter, num[jj], self.h.level.spawns[jj])

            # the pairs of cyans from the end of wave 3, but twice as often!
            if p > 2/3 and (self.ticks - 10) % int(1800 / d) == 0:
                self.add_group(entity.Seeker, 2)
        
        elif self.n == 9:

            # wave 9 begins with two rings of oranges, then transitions to a bombardment of
            # random enemies
            
            if self.ticks == 1 or self.ticks == int(WAVE_TICKS / 3):
                num = [0] * 8
                for ii in xrange(int(d / 3)):
                    num[self.rng.randrange(8)] += 1
                for jj in xrange(8):    
                    self.add_group_at_pos(entity.Birther, num[jj], self.h.level.spawns[jj])

            if p > 1/7 and (self.ticks - 80) % int(1500 / d) == 0:
                cls = self.rng.choice([entity.Bumbler, entity.BumblingSeeker, entity.Whirly, entity.Robot, entity.Darter])
                self.add_group(cls, self.rng.randrange(4))
                
            if p > 1/2 and (self.ticks - 90) % int(3500 / d) == 0:
                self.add_group(entity.Sniper, 2)

            if p > 3/4 and (self.ticks - 100) % int(3000 / d) == 0:
                self.add_group(entity.Breeder, 2)
                
            if self.ticks > WAVE_TICKS - 3.5 * TICK_RATE and (self.ticks % int(50 / d)) == 0:
                self.add_group(entity.Seeker, 1)
        
    def add_group(self, cls, number):
        self.add_group_at_pos(cls, number, self.rng.choice(self.h.level.spawns))

    def add_group_at_pos(self, cls, number, pos):
        disp = 0 if number == 1 else 20
        for ii in xrange(number):
            self.new.append((cls, pos + v((disp,0)).rotated(self.rng.randrange(360))))
        self.count += number
            
    def new_enemies(self):
        res = []
        for cls, pos in self.new:
            res.append((cls, pos, self.seed, self.echo))
            self.seed += 1
        self.new = []
        return res
